
//** Modified-Light version of Connectivity for TP **//


#ifndef   	CONNECTIVITY_H_
# define   	CONNECTIVITY_H_

#include <stdlib.h>
#include <iostream>
#include <cmath>
#include <string.h>
#include <string>
#include <errno.h>
#include <stdio.h>
#include <int_vector.h>

#include <vector>
#include <algorithm>

#include <Point_set.h>


/// Connectivity of a mesh
/** Container class for the connectivity

    Separate the general connectivity (any polygons), but also store a
    triangulated version for the openGl draw

    Save general polygon in an array with separator=-1 (as in VRML):
    [k_1,k_2,k_3,k_4,-1,k_5,k_6,k_7,-1,...]
    also know polygon size:
    [4,3,...]
    and a direct access index:
    [0,5,...]

    Also save a triangulated version for fast drawing.

*/
class Connectivity
{
 public:
  
  //************************************************//
  //************************************************//
  //CONSTRUCTORS
  //************************************************//
  //************************************************//

  /// classical constructor
  Connectivity();
  /// Copy constructor
  Connectivity(const Connectivity& );
  /// classical destructor
  ~Connectivity();
  /// Set all size to 0
  int destroy();





  //************************************************//
  //************************************************//
  //Access/Size
  //************************************************//
  //************************************************//

  /// get index from k_index
  int get_index(int k_index) const;
  /// get index from k_index
  int& get_index(int k_index);
  /// get index from k_index of k_polygon
  int get_index(int k_polygon,int k_index) const;
  /// get index from k_index of k_polygon
  int& get_index(int k_polygon,int k_index);
  /// get full_index of the given polygon
  std::vector <int> get_index_of_polygon(int k_polygon) const;
  /// get_pointer
  const int* get_index() const;


  /// size of the vector
  int size_index() const;
  /// number of polygons
  int polygon_number() const;
  /// get size of the given polygon
  int polygon_size(int k_polygon) const;
  /// get number of vertex by taking the max of every vertex index
  int vertex_number() const;


  /// get index of the triangulated version
  int get_index_triangulated(int k_index) const;
  /// get index of the triangulated version
  int& get_index_triangulated(int k_index);
  /// get index of this triangle in the triangulated version
  int get_index_triangulated(int k_triangle,int k_index) const;
  /// get index of this triangle in the triangulated version
  int& get_index_triangulated(int k_triangle,int k_index);
  /// get pointer of the triangulation
  const int* get_index_triangulated() const;
  /// get number of triangules in the triangulated version
  int get_triangle_number() const;

  /// triangulate the mesh
  /** Suppose that the triangulated part is computed*/
  int triangulate();

  /// try to find a given polygon return polygon number or -1 if not found
  int find_polygon(const std::vector <int>& polygon) const;


  /// get the polygon access (when polygons are given in the same order )
  /**
     index0,index1,index2,-1,index3,index4,...
   */
  int get_polygon_access(int k_polygon) const;


  //************************************************//
  //************************************************//
  //Build access
  //************************************************//
  //************************************************//

  /// fill the array of polygon size and the direct access polygon
  int build_size_and_polygon_access();
  /// build the triangulated version of the index
  /** Perform a triangle strip build for each polygon */
  int build_triangulated_index();
  /// find max_index to estimate the number of vertex
  int find_max_index();
  /// build all starting from index_polygon only
  /** actually call build_size_and_polygon_access() and
      build_triangulated_index() functions */
  int build_all_index();
  








  //************************************************//
  //************************************************//
  //Add/Delete polygon
  //************************************************//
  //************************************************//

  /// add a full polygon
  int add_polygon(const std::vector <int>& polygon);
  /// add a full polygon
  int add_polygon(const int_vector& polygon);
  /// add a full quad
  int add_polygon(int index_0,int index_1,int index_2,int index_3);
  /// add a full triangle
  int add_polygon(int index_0,int index_1,int index_2);
  /// add index of polygon
  /** does not ensure that the polygon end by -1 (should be manually included)*/
  int add_index(int index);
  /// set polygon number k
  /** slow, needs the re-calculation of the index 
      Also assume that the access is already set-up */
  int set_polygon(int k_polygon,const std::vector <int>& polygon);
  /// set index
  /** should not be use, direct access to the index, does not ensure
      its correct with size */
  int set_index(int k_polygon,int k_index,int index);
  /// delete polygon
  /** assume that all the index are already calculated.
      Slow, will recalculate everything at each time */
  int delete_polygon(int k_polygon);
  /// delete polygon from a given polygon
  int delete_given_polygon(const std::vector <int> polygon);

  
  /// Flip a given polygon
  /** Inverse the sens of the polygon */
  int flip_polygon(int k_polygon);
  /// Flip every polygons
  int flip_polygon();



  //************************************************//
  //************************************************//
  //One Ring
  //************************************************//
  //************************************************//

  /// build one ring on a general mesh (not necessarily 2-manifold)
  /** Also build the polygon ring */
  int build_one_ring();

  /// get one ring
  int_vector get_one_ring(int k_polygon,int k_edge) const;
  /// get one ring directly from the vertex
  int_vector get_one_ring(int k_vertex) const;

  /// get polygon ring (polygon that shares this vertex)
  int_vector get_polygon_ring(int k_polygon,int k_edge) const;
  /// get polygon ring directly form the vertex (polygon that shares this vertex)
  int_vector get_polygon_ring(int k_vertex) const;



  //************************************************//
  //************************************************//
  //INPUT/OUTPUT
  //************************************************//
  //************************************************//


  /// load off connectivity
  int load_off_file(const char* filename);
  /// load off connectivity
  int load_off_file(const std::string& filename);
  /// load obj connectivity
  int load_obj_file(const char* filename);
  /// load obj connectivity
  int load_obj_file(const std::string& filename);

  /// load g file connectivity
  int load_g_file(const char* filename);
  /// load g file connectivity
  int load_g_file(const std::string& filename);
  /// write g type file connectivity
  int write_g_type(const char* filename) const;
  /// write g type file connectivity
  int write_g_type(const std::string& filename) const;


  /// Add connectivity description in a .off type
  int write_off_type(const char* filename) const;
  /// Add connectivity description in a .off type
  int write_off_type(const std::string& filename) const;
  /// Add connectivity description in a .obj type
  int write_obj_type(const char* filename) const;
  /// Add connectivity description in a .obj type
  int write_obj_type(const std::string& filename) const;



  //************************************************//
  //************************************************//
  //OPERATORS
  //************************************************//
  //************************************************//

  /// direct access to index_polygon
  int operator()(int k_index) const;
  /// direct access to index_polygon
  int operator[](int k_index) const;
  /// direct access to index_polygon
  int& operator()(int k_index);
  /// direct access to index_polygon
  int& operator[](int k_index);
  /// access to k_index of polygon n. k_polygon
  int operator()(int k_polygon,int k_index) const;
  /// access to k_index of polygon n. k_polygon
  int& operator()(int k_polygon,int k_index);

  //affectation operator
  const Connectivity& operator=(const Connectivity&);

  /// concatenation of two connectivity
  /** Just add, assume there is no duplicate */
  friend Connectivity operator+(const Connectivity&,const Connectivity&);
  /// add polygon
  friend Connectivity operator+(const Connectivity&,const std::vector<int> &);
  /// add polygon
  friend Connectivity operator+(const std::vector<int> &,const Connectivity&);
  /// delete a set of polygon
  friend Connectivity operator-(const Connectivity&,const Connectivity&);
  /// delete polygon
  friend Connectivity operator-(const Connectivity&,const std::vector<int>&);
  /// test equality
  friend bool operator==(const Connectivity&,const Connectivity&);
  /// test equality
  friend bool operator!=(const Connectivity&,const Connectivity&);
  /// Stream output
  friend ostream& operator << (ostream& flux,const Connectivity& );





 private:


  /// look to find the edge corresponding to [vertex_0,vertex_1],
  /// return -1 if not
  /** Full research, slow. Only use, when access table is not build */
  int find_current_edge_slow(int vertex_index_0,int vertex_index_1);
  /// look to find all the edge corresponding to [vertex_0,vertex_1]
  /// return empty vector if not found
  /** Full research, slow. */
  std::vector <int> find_all_current_edge_slow(int vertex_index_0,int vertex_index_1);

  /// Pass through all the edges to flip the polygon if needed
  /** Reccursive construction */
  int reccursive_connect_manifold_polygons(int current_polygon,std::vector <int>* correct_polygon);




  /// every index are ordered in only one vector
  /** vector of type [index_1,index_2,index_3,-1,index_4,...]. Here -1
      is used as a separator. */
  std::vector <int> index_polygon;

  /// max_index estimate the number of vertex that are needed
  /*** could be better using an indirection. Just give the max index
       through all the index*/
  int max_index;

  /// vertex of polygon size
  std::vector <int> size_polygon;

  /// direct access of polygon n.k
  std::vector <int> polygon_access;

  /// triangulated version
  std::vector <int> triangulated_index;

  /// one_ring
  std::vector <int_vector> one_ring;

  /// Polygon ring = polygons surrounding a given vertex
  std::vector <int_vector> polygon_ring;

  /// Set of Edges of the mesh
  // std::vector <Edge> edges;
  
  /// Pointer between polygon edges and index in Edge vector
  std::vector <int> index_edges;

  /// Pointer between [index_0,index_1] and edge index for direct access
  std::vector <int_vector> edges_access;

};

#endif 	    /* !CONNECTIVITY_H_ */
